/*
** trace.c for elfsh
**
** All functions that help to trace content
**
** Started Jul 2 2005 00:03:44 mxatone
**
**
** $Id: trace.c,v 1.11 2007/03/08 20:50:21 mxatone Exp $
**
*/
#include "revm.h"

char buf[BUFSIZ];

/* XXX: The good syntax should be : 
** 
** trace <add>     funcname     [optional_traces_name]  : add a function to a given trace
** trace <rm>      funcname     [optional_traces_name]  : remove a function from a given trace
** trace <enable>  funcname|all [optional_traces_name]  : enable tracing for a function in a given trace (enabled by default after a add)
** trace <disable> funcname|all [optional_traces_name]  : disable tracing for a function in a given trace
** trace <create>  tracename    <optionals funcnames>  : create a new trace with a pool of traced functions by default
** trace <delete>  tracename                           : delete a trace
** trace <flush>   tracename                           : remove all functions from a given trace
** trace <list>                                        : list all available traces
**
*/                             

/**
 * Research the uniform debugging format function entry on a list of files 
 * @param files first element of a edfmtfile_t structure
 * @param name function name
 * @return pointer on a edfmtfunc_t (represent function debugging information)
 */
static edfmtfunc_t	*elfsh_traces_search_uni_file(edfmtfile_t *files, char *name)
{
  edfmtfunc_t		*func;
  edfmtfile_t		*file;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  for (file = files ; file != NULL; file = file->next)
    {
      func = hash_get(&(file->hfunc), name);

      if (func != NULL)
	PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, func); 

      if (file->child)
	{
	  func = elfsh_traces_search_uni_file(file->child, name);
	  if (func != NULL)
	    PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, func); 
	}
    }

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, NULL); 
}

/**
 * Research the uniform debugging format function entry 
 * @param file first element of a edfmtfile_t structure
 * @param name function name
 * @return pointer on a edfmtfunc_t (represent function debugging information)
 */
static edfmtfunc_t    	*elfsh_traces_search_uni(elfshobj_t *file, char *name)
{
  edfmtfunc_t		*func;
  edfmtinfo_t		*uni;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  uni = edfmt_get_uniinfo(file);
 
  if (!uni)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "No debugging informations available", NULL); 

  /* Search on global scope */
  func = hash_get(&(uni->hfunc), name);
  if (func == NULL)
    {
      /* Search on local stop if needed */
      func = elfsh_traces_search_uni_file(uni->files, name);
      if (func == NULL)
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		     "Function not found", NULL); 
    }

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, func); 
}

/**
 * Search the function entry on the uniform debugging format, and return it 
 * @param file first element of a edfmtfile_t structure
 * @param func_name function name
 * @param external indicate if we need internal or external research
 * @return pointer on a edfmtfunc_t (represent function debugging information)
 */
edfmtfunc_t 		*elfsh_traces_tracable_with_type(elfshobj_t *file, char *func_name,
							 u_char external)
{
  elfshobj_t		*sym_file;
  edfmtfunc_t		*func = NULL;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  /* We have an extern symbol */
  if (external)
    {
      sym_file = elfsh_traces_search_sym(file, func_name);
      
      if (sym_file)
	func = elfsh_traces_search_uni(sym_file, func_name);
    }
  else
    {
      /* Search in this file */
      func = elfsh_traces_search_uni(file, func_name);
    }

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, func);
}

/**
 * Construct the argument structure depending of the option we have 
 * @param file target file
 * @param ent pointer to set
 * @param func uniform debugging inforation if we got one
 * @param vaddr virtual address of the function if we don't have debugging information
 * @param external describe scope
 * @see elfsh_traces_createargs
 */
static int		elfsh_traces_build_args(elfshobj_t *file, elfshtraces_t *ent, 
						edfmtfunc_t *func, elfsh_Addr vaddr, 
						u_char external)
{
  u_int			index;
  int			*argcount;
  edfmtfuncarg_t	*arg;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__); 

  if (!ent)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "Invalid parameters", -1);

  ent->scope = external ? ELFSH_ARG_INTERN : ELFSH_ARG_INTERN;

  /* Do we have uniform debugging information for this structure ? */
  if (func == NULL)
    {
      argcount = elfsh_args_count(file, 0, vaddr);
      
      if (argcount == NULL)
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		     "Arg count failed", -1);

      ent->type = ELFSH_ARG_SIZE_BASED;
      
      /* Fill arguments */
      for (index = 0; argcount[index] != NULL; index++)
	ent->arguments[index].size = argcount[index];

      ent->argc = index;
      XFREE(__FILE__, __FUNCTION__, __LINE__, argcount);
    }
  else
    {
      ent->type = ELFSH_ARG_TYPE_BASED;
      for (arg = func->arguments, index = 0; arg != NULL; arg = arg->next, index++)
	{
	  /* We setup argument name and typename */
	  ent->arguments[index].name = arg->name;
	  ent->arguments[index].typename = arg->type ? arg->type->name : NULL;

	  /* Other information are quite the same as arg counting method */
	  if (arg->next == NULL)
	    {
	      if (arg->type->size > 0)
		ent->arguments[index].size = arg->type->size; 
	      else
		ent->arguments[index].size = sizeof(void *); 
	    }
	  else
	    {
	      ent->arguments[index].size = arg->next->pos - arg->pos; 
	    }
	}

      ent->argc = func->argc;
    }
      
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

/**
 * Allocate the traces  structure and call previous
 * @param file target file
 * @param func uniform debugging inforation if we got one
 * @param vaddr virtual address of the function if we don't have debugging information
 * @param external describe scope
 * @return generate trace structure
 * @see elfsh_traces_build_args
 */
elfshtraces_t		*elfsh_traces_createargs(elfshobj_t *file, char *name,
						 edfmtfunc_t *func, elfsh_Addr vaddr,
						 u_char external)
{
  elfshtraces_t 	*newtrace;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);
 

  // Alloc the structure
  XALLOC(__FILE__, __FUNCTION__, __LINE__, newtrace, sizeof(elfshtraces_t), NULL);
  newtrace->funcname = strdup(name);
  newtrace->file = file;
  newtrace->enable = 1;

  // We fill arguments part and check if we found the function
  if (elfsh_traces_build_args(file, newtrace, func, vaddr, external) < 0)
    {
      XFREE(__FILE__, __FUNCTION__, __LINE__, newtrace->funcname);
      XFREE(__FILE__, __FUNCTION__, __LINE__, newtrace);
      PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		   "Symbol not found or impossible to trace", NULL);
     } 

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, newtrace);
}

/**
 * Add a fonction on a trace (from commande line)
 * @param name function name
 * @param optarg trace name (optional)
 */
int		traces_add(char *name, char **optarg)
{
  elfshtraces_t *traces;
  edfmtfunc_t	*func = NULL;
  elfsh_Addr	vaddr = 0;
  u_char	external = 0;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  if (!name || !name[0])
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "Invalid parameters", -1);
  
  /* Can we trace this function */
  if (elfsh_traces_tracable(world.curjob->current, name, &vaddr, &external) < 0)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "Untracable function", -1);

  /* Can we trace with libedfmt informations ? */
  func = elfsh_traces_tracable_with_type(world.curjob->current, name, external);

  /* Create the structure */
  traces = elfsh_traces_createargs(world.curjob->current, name, func, vaddr, external);

  if (!traces)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "Failed to create trace arguments", -1);
  
  /* Add the function */
  if (elfsh_traces_funcadd(optarg ? *optarg : NULL, name, traces) == NULL)
    {
      /* XFREE here ? */
      PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		   "Add function failed", -1);
    }

  snprintf(buf, BUFSIZ - 1, "\t[*] Added function %s successfully into trace %s\n\n",
	   name, optarg && *optarg ? *optarg : ELFSH_TRACES_TYPE_DEFAULT);
  vm_output(buf);

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}
 
/** 
 * Delete a function from a trace 
 * @param name function name
 * @param optarg trace name (optional)
 */
int		traces_rm(char *name, char **optarg)
{
  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  if (!name || !name[0])
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "Invalid parameters", -1);

  if (elfsh_traces_funcrm(optarg ? *optarg : NULL, name) < 0)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "Delete function failed", -1);

  snprintf(buf, BUFSIZ - 1, "\t[*] Deleted function %s successfully from trace %s\n\n",
	   name, optarg && *optarg ? *optarg : ELFSH_TRACES_TYPE_DEFAULT);
  vm_output(buf);

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

/**
 * Enable a function from a trace 
 * @param name function name
 * @param optarg trace name (optional)
 */
int		traces_enable(char *name, char **optarg)
{
  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  if (!strcmp(name, "all"))
    {
      if (elfsh_traces_funcenableall(optarg ? *optarg : NULL) < 0)
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		     "Enable all functions failed", -1);      
    }
  else
    {
      if (elfsh_traces_funcenable(optarg ? *optarg : NULL, name) < 0)
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		     "Enable function failed", -1);
    }

  snprintf(buf, BUFSIZ - 1, "\t[*] Enabled function %s successfully from trace %s\n\n",
	   name, optarg && *optarg ? *optarg : ELFSH_TRACES_TYPE_DEFAULT);
  vm_output(buf);

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

/**
 * Disable a function from a trace 
 * @param name function name
 * @param optarg trace name (optional)
 */
int		traces_disable(char *name, char **optarg)
{
  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  if (!strcmp(name, "all"))
    {
      if (elfsh_traces_funcdisableall(optarg ? *optarg : NULL) < 0)
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		     "Disable all functions failed", -1);      
    }
  else
    {
      if (elfsh_traces_funcdisable(optarg ? *optarg : NULL, name) < 0)
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		     "Disable function failed", -1);
    }

  snprintf(buf, BUFSIZ - 1, "\t[*] Disabled function %s successfully from trace %s\n\n",
	   name, optarg && *optarg ? *optarg : ELFSH_TRACES_TYPE_DEFAULT);
  vm_output(buf);

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

/**
 * Create a trace 
 * @param name trace name
 * @param optarg list of function to directly add
 */
int		traces_create(char *name, char **optarg)
{
  char		*strarr[2];
  int		index;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  if (!name || !name[0])
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "Invalid parameters", -1);

  if (elfsh_traces_createtrace(name) == NULL)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "Create trace failed", -1);

  snprintf(buf, BUFSIZ - 1, "\t[*] Created trace %s successfully\n\n", name);
  vm_output(buf);

  if (optarg && *optarg)
    {
      /* Add gived elements */
      for (index = 0; optarg[index]; index++)
	{
	  strarr[0] = optarg[index];
	  strarr[1] = NULL;
	  if (traces_add(name, strarr) < 0)
	    continue;
	  
	  snprintf(buf, BUFSIZ - 1, "\t[*] Added function %s successfully\n", optarg[index]);
	  vm_output(buf);	  
	}

      vm_output("\n");
    }
    
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

/**
 * Delete a trace 
 * @param name trace name
 * @param optarg not use
 */
int		traces_delete(char *name, char **optarg)
{
  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  if (!name || !name[0])
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "Invalid parameters", -1);

  if (elfsh_traces_deletetrace(name) < 0)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "Delete trace failed", -1);

  snprintf(buf, BUFSIZ - 1, "\t[*] Deleted trace %s successfully\n\n", name);
  vm_output(buf);

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

/**
 * Flush a whole trace 
 * @param name trace name
 * @param optarg not use
 */
int		traces_flush(char *name, char **optarg)
{
  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  if (!name || !name[0])
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "Invalid parameters", -1);

  if (elfsh_traces_funcrmall(name) < 0)
    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__, 
		 "Flush trace failed", -1);

  snprintf(buf, BUFSIZ - 1, "\t[*] Flushed trace %s successfully\n\n", name);
  vm_output(buf);

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

/**
 * Detail listing of a trace 
 * @param table trace hash table
 * @param name trace name
 * @see trace_list
 */
static int	traces_list_detail(hash_t *table, char *name)
{
  char		**keys;
  u_int		index;
  int		keynbr;
  elfshtraces_t	*entrie;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  snprintf(buf, BUFSIZ - 1, " - Trace name: %s -\n\n", name);
  vm_output(buf);

  keys = hash_get_keys(table, &keynbr);

  if (keys)
    {
      for (index = 0; index < keynbr; index++)
	{
	  entrie = (elfshtraces_t *) hash_get(table, keys[index]);
	  
	  if (entrie)
	    {
	      snprintf(buf, BUFSIZ - 1, " Function name: %-15s status: %s\n",
		       entrie->funcname, entrie->enable ? "enabled" : "disabled");
	      vm_output(buf);
	    }
	}

      vm_output("\n");

      hash_free_keys(keys);
    }

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

/**
 * Enumerate each trace and detail every functions 
 * @param name not use
 * @param optarg not use
 * @see traces_list_detail
 */
int		traces_list(char *name, char **optarg)
{
  char		**keys;
  u_int		index;
  int		keynbr;
  hash_t	*subtable;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  vm_output(" .: Trace list :.\n");

  keys = hash_get_keys(&traces_table, &keynbr);

  if (keys)
    {
      for (index = 0; index < keynbr; index++)
	{
	  subtable = hash_get(&traces_table, keys[index]);

	  if (subtable)
	    traces_list_detail(subtable, keys[index]);
	}

      hash_free_keys(keys);
    }
  else
    {
      vm_output("No table available\n");
    }

  vm_output("\n");

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

/**
 * Create the structure that store trace cmd informations 
 * @param exec function pointer
 * @param flagName
 * @param flagArg
 */
revmtraces_t   *traces_create_CMDENT(int (*exec)(char *name, char **optarg),
				   char flagName, char flagArg)
{
  revmtraces_t *new;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  XALLOC(__FILE__, __FUNCTION__, __LINE__, new, sizeof(revmtraces_t), NULL);
  new->exec	= exec;
  new->flagName	= flagName;
  new->flagArg	= flagArg;
 
  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, (new));
}

/**
 * Add a command in sub commande hash table
 * @param cmd cmd name
 * @param exec function pointer
 * @param flagName
 * @param flagArg
 * @see traces_cmd_hash
 * @see traces_create_CMDENT
 */
int 		traces_addcmd(char *cmd, void *exec, char flagName, char flagArg)
{
  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);

  hash_add(&traces_cmd_hash, cmd, (void *) traces_create_CMDENT(exec, flagName, flagArg));

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, 0);
}

int 			cmd_traces()
{
  revmtraces_t		*cmd;
  char			*fArg;
  char			**sArg;
  int			ret = 0;
  hash_t		*table;

  PROFILER_IN(__FILE__, __FUNCTION__, __LINE__);
  
  /* Init tracing elements */
  elfsh_traces_inittrace();

  switch(world.curjob->curcmd->argc)
    {
    case 0:
      traces_list(NULL, NULL);
      break;
    case 1:
      table = elfsh_traces_gettrace(world.curjob->curcmd->param[0]);

      if (!table)
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
		     "Unknown trace", -1);

      traces_list_detail(table, world.curjob->curcmd->param[0]);
      break;
    default:
      cmd = hash_get(&traces_cmd_hash, world.curjob->curcmd->param[0]);
      
      if (!cmd)
	PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
			  "Unknown command", -1);

      fArg = NULL;
      sArg = NULL;

      if (cmd->flagName > 0)
	{
	  /* First argument is needed and doesn't submited */
	  if (!world.curjob->curcmd->param[1] && cmd->flagName == 2)
	    PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
			      "First argument is needed", -1);

	  fArg = world.curjob->curcmd->param[1];

	  if (cmd->flagArg > 0)
	    {
	      /* Second argument is needed and doesn't submited */
	      if (!world.curjob->curcmd->param[2] && cmd->flagArg == 2)
		PROFILER_ERR(__FILE__, __FUNCTION__, __LINE__,
				  "Second argument is needed", -1);

	      sArg = world.curjob->curcmd->param + 2;
	    }
	}

      ret = cmd->exec(fArg, sArg);      
    }

  PROFILER_ROUT(__FILE__, __FUNCTION__, __LINE__, (ret));
}
